------------------------- MODULE DistributedCommit -------------------------
(***************************************************************************)
(* This is a high-level TLA+ specification of a distributed commit         *)
(* protocol. A set of nodes individually decide whether they wish to       *)
(* commit or abort a transaction. Second, the transaction will be          *)
(* committed iff every node wishes to commit.                              *)
(***************************************************************************)
CONSTANT Node  \* set of participating nodes

VARIABLE nState  \* state of each participant

(***************************************************************************)
(* Possible states of nodes.                                               *)
(***************************************************************************)
NState == {"preparing", "readyCommit", "readyAbort", "committed", "aborted"}

(***************************************************************************)
(* Initially, every node is preparing the transaction.                     *)
(***************************************************************************)
Init == nState = [n \in Node |-> "preparing"]

(***************************************************************************)
(* A node decides whether it wishes to commit or abort.                    *)
(***************************************************************************)
Decide(n) ==
  /\ nState[n] = "preparing"
  /\ \/ nState' = [nState EXCEPT ![n] = "readyCommit"]
     \/ nState' = [nState EXCEPT ![n] = "readyAbort"]

(***************************************************************************)
(* A node may commit only if all nodes wish to do so.                      *)
(***************************************************************************)
Commit(n) ==
  /\ \A q \in Node : nState[q] \in {"readyCommit", "committed"}
  /\ nState' = [nState EXCEPT ![n] = "committed"]

(***************************************************************************)
(* A node aborts if some node requests an abort.                           *)
(***************************************************************************)
Abort(n) ==
  /\ \E q \in Node : nState[q] \in {"readyAbort", "aborted"}
  /\ nState' = [nState EXCEPT ![n] = "aborted"]

(***************************************************************************)
(* The next-state relation is the disjunction of the above actions.        *)
(***************************************************************************)
Next == \E n \in Node : Decide(n) \/ Commit(n) \/ Abort(n)

(***************************************************************************)
(* Overall specification.                                                  *)
(***************************************************************************)
Spec == Init /\ [][Next]_nState  /\ WF_nState(Next)

-----------------------------------------------------------------------------
(***************************************************************************)
(* Check type correctness.                                                 *)
(***************************************************************************)
TypeOK == nState \in [Node -> NState]

(***************************************************************************)
(* Main safety property: nodes may commit only if all nodes agree.         *)
(***************************************************************************)
Safety == 
  \A n \in Node : nState[n] = "committed" =>
     \A q \in Node : nState[n] \in {"readyCommit", "committed"}

(***************************************************************************)
(* Non-property: no node may ever commit.                                 *)
(***************************************************************************)
NeverCommit == \A n \in Node : nState[n] # "committed"

(***************************************************************************)
(* Liveness property: nodes will eventually commit or abort.               *)
(* (Requires fairness condition on next-state relation.)                   *)
(***************************************************************************)
Liveness == \A n \in Node : <>(nState[n] \in {"committed", "aborted"})

=============================================================================
