--------------------------- MODULE TwoPhaseCommit ---------------------------
(***************************************************************************)
(* This is a TLA+ specification of the two-phase commit protocol used for  *)
(* distributed data bases. It models only one instance of the protocol,    *)
(* i.e. for a single transaction.                                          *)
(***************************************************************************)

CONSTANT Node  \* set of nodes other than the coordinator

VARIABLES
  cState,     \* the state of the coordinator
  nState,     \* the state of the non-coordinator nodes
  committed,  \* nodes that the coordinator knows are OK for committing
  msgs        \* messages sent during the protocol

vars == <<cState, nState, committed, msgs>>

(***************************************************************************)
(* Possible states of coordinator.                                         *)
(***************************************************************************)
CState == {"preparing", "committed", "aborted"}

(***************************************************************************)
(* Possible states of non-coordinator participants.                        *)
(***************************************************************************)
NState == {"preparing", "readyCommit", "readyAbort", "committed", "aborted"}

(***************************************************************************)
(* Messages sent during the protocol.                                      *)
(***************************************************************************)
Message ==
  \* node informs coordinator about its decision
  [kind : {"commit", "abort"}, node : Node] \cup
  \* coordinator tells nodes whether to commit or abort
  [kind : {"doCommit", "doAbort"}]

commit(n) == [kind |-> "commit", node |-> n]
abort(n) == [kind |-> "abort", node |-> n]
doCommit == [kind |-> "doCommit"]
doAbort == [kind |-> "doAbort"]

(***************************************************************************)
(* The following predicate specifies what values the variables can take    *)
(* during an execution of the protocol.                                    *)
(***************************************************************************)
TypeOK ==
  /\ cState \in CState
  /\ nState \in [Node -> NState]
  /\ committed \subseteq Node
  /\ msgs \subseteq Message
  
(***************************************************************************)
(* The initial state of the protocol.                                      *)
(***************************************************************************)
Init ==
  /\ cState = "preparing"
  /\ nState = [n \in Node |-> "preparing"]
  /\ committed = {}
  /\ msgs = {}

(***************************************************************************)
(* The following action formulas describe the possible transitions of      *)
(* the nodes.                                                              *)
(***************************************************************************)

(***************************************************************************)
(* A participant decides and informs the coordinator of its decision.      *)
(***************************************************************************)
Decide(n) ==
  /\ nState[n] = "preparing"
  /\ \/ /\ nState' = [nState EXCEPT ![n] = "readyCommit"]
        /\ msgs' = msgs \cup {commit(n)}
     \/ /\ nState' = [nState EXCEPT ![n] = "readyAbort"] 
        /\ msgs' = msgs \cup {abort(n)}
  /\ UNCHANGED <<cState, committed>>

(***************************************************************************)
(* The coordinator receives a new commit decision for some participant.    *)
(* If all participants wish to commit, it sends an order to commit.        *)
(***************************************************************************)
RcvCommit(n) == 
  /\ n \notin committed /\ commit(n) \in msgs
  /\ committed' = committed \cup {n}
  /\ IF committed' = Node
     THEN /\ cState' = "committed"
          /\ msgs' = msgs \cup {doCommit}
     ELSE UNCHANGED <<cState, msgs>> 
  /\ nState' = nState

(***************************************************************************)
(* The coordinator receives an abort decision and sends an order to abort. *)
(***************************************************************************)
RcvAbort(n) == 
  /\ abort(n) \in msgs
  /\ cState' = "aborted"
  /\ msgs' = msgs \cup {doAbort}
  /\ UNCHANGED <<nState, committed>>

(***************************************************************************)
(* A participant receives a commit or abort order from the coordinator.    *)
(***************************************************************************)
Execute(n) ==
  /\ \/ doCommit \in msgs /\ nState' = [nState EXCEPT ![n] = "committed"]
     \/ doAbort \in msgs /\ nState' = [nState EXCEPT ![n] = "aborted"]
  /\ UNCHANGED <<cState, committed, msgs>>

(***************************************************************************)
(* The overall next-state relation is the disjunction of the action        *)
(* formulas defined previously.                                            *)
(***************************************************************************)
Next == 
  \E n \in Node : Decide(n) \/ RcvCommit(n) \/ RcvAbort(n) \/ Execute(n)

Spec == Init /\ [][Next]_vars /\ WF_vars(Next)
-----------------------------------------------------------------------------

NoConflictingOrders == 
  doCommit \in msgs =>
    /\ ~(doAbort \in msgs)
    /\ \A q \in Node : nState[q] \in {"readyCommit", "committed"}

(***************************************************************************)
(* Correctness properties.                                                 *)
(***************************************************************************)

(***************************************************************************)
(* The coordinator never sends both a doCommit and a doAbort message.      *)
(***************************************************************************)
CommitOrAbort == ~(doCommit \in msgs /\ doAbort \in msgs)

(***************************************************************************)
(* The coordinator may commit only if all participants wish to commit and  *)
(* no participant wishes to abort.                                         *)
(***************************************************************************)
AbortWins == 
  doCommit \in msgs => 
    \A n \in Node : 
       /\ commit(n) \in msgs /\ nState[n] \in {"readyCommit", "committed"}
       /\ abort(n) \notin msgs

Terminal ==
  \/ cState = "aborted" /\ \A n \in Node : nState[n] = "aborted"
  \/ cState = "committed" /\ \A n \in Node : nState[n] = "committed"

(***************************************************************************)
(* The protocol may only terminate in a terminal state.                    *)
(***************************************************************************)
CorrectTermination == (~ ENABLED <<Next>>_vars) => Terminal

Liveness == <>Terminal
-----------------------------------------------------------------------------

(***************************************************************************)
(* Two-phase commitment implements distributed commitment.                 *)
(***************************************************************************)
DC == INSTANCE DistributedCommit

THEOREM Spec => DC!Spec
=============================================================================
